from struct import unpack
import numpy as np
import scipy.io as sio
import os
import re
import stmpy
from stmpy import matio
from datetime import datetime, timedelta
from scipy.optimize import minimize

'''
I think I should use more doc strings.
'''


def load(filePath, biasOffset=True, niceUnits=False):
    '''
    Loads data into python. Please include the file extension in the path.

    Currently supports formats: 3ds, sxm, dat, nvi, nvl, mat.

    For .3ds and .dat file types there is an optional flag to correct for bias offset
    that is true by default.  This does not correct for a current offset, and
    should not be used in cases where there is a significant current offset.
    Note: .mat files are supported as exports from STMView only.

    Inputs:
        filePath    - Required : Path to file including extension.
        baisOffset  - Optional : Corrects didv data for bias offset by looking
                                 for where the current is zero.
        niceUnits   - Optional : Put lock-in channel units as nS (in future
                                 will switch Z to pm, etc.)
    Returns:
        dataObject  - Custom object with attributes appropriate to the type of
                      data and containing experiment parameters in a header.
                      
    Usage:
        data = load('file.3ds', biasOffset=True, niceUnits=False)
    '''
    if filePath.endswith('.3ds'):
        dataObject = Nanonis3ds(filePath)
        if biasOffset:
            dataObject = _correct_bias_offset(dataObject, '.3ds')
        if niceUnits:
            dataObject = _nice_units(dataObject)
        return dataObject

    elif filePath.endswith('.sxm'):
        return NanonisSXM(filePath)

    elif filePath.endswith('.dat'):
        if biasOffset:
            return _correct_bias_offset(NanonisDat(filePath), '.dat')
        else:
            return NanonisDat(filePath)

    elif filePath[-3:] == 'NVI' or filePath[-3:] == 'nvi':
        return NISTnvi(sio.readsav(filePath))

    elif filePath[-3:] == 'NVL' or filePath[-3:] == 'nvl':
        return NISTnvl(sio.readsav(filePath))

    elif filePath.endswith('.nsp'):
        return LongTermSpectrum(filePath)

    elif filePath.endswith('.mat'):
        raw_mat = matio.loadmat(filePath)
        mappy_dict = {}
        for key in raw_mat:
            try:
                mappy_dict[key] = matio.Mappy()
                mappy_dict[key].mat2mappy(raw_mat[key])
                print('Created channel: {:}'.format(key))
            except:
                del mappy_dict[key]
                print('Could not convert: {:}'.format(key))
        if len(mappy_dict) == 1: return mappy_dict[mappy_dict.keys()[0]]
        else: return mappy_dict
    
    elif filePath.endswith('.asc'):
        return AsciiFile(filePath)

    else: raise IOError('ERR - Wrong file type.')


def save(filePath, pyObject):
    '''
Save objects from a python workspace to disk.
Currently implemented for the following python data types: nvl, mat.
Currently saves to the following file types: mat.
Please include the file extension in the path, e.g. 'file.mat'

Usage: save(filePath, data)
    '''
    if filePath.endswith('.mat'):
        if pyObject.__class__ == matio.Mappy:
            pyObject.savemat(filePath)
        elif pyObject.__class__ == NISTnvl:
            mappyObject = matio.Mappy()
            mappyObject.nvl2mappy(pyObject)
            mappyObject.savemat(filePath)
    else: raise IOError('ERR - File format not supported.')

def qkrdasciifile(filename, delimiter='\t', returnnotes=False):
    '''
    Read formatted data in a general ascii file (no specific extension) with header and endnote optionally returned.
    Data is read when a line starts with a digit (leading spaces and tabs skipped), other than which all above saved as header, below saved as endnote, and both printed out.
    Usage: data, header, endnote = qkrdasciifile('filename.txt', delimiter='\t', returnnotes=True)
    '''
    with open(filename, 'r') as f:
        lines = f.readlines()
    header = []
    ih = 0
    while not lines[ih].lstrip('- \t')[0].isdigit():
        ih = ih + 1
    header = lines[:ih]
    idt = 0
    while lines[ih+idt].lstrip('- \t')[0].isdigit():
        idt = idt + 1
        if ih+idt == len(lines):
            break
    endnote = lines[(ih+idt):]
    rows = idt
    cols = len(lines[ih].lstrip('\t').split(delimiter))
    data = np.zeros((cols, rows))
    for ir in range(rows):
        tmp = lines[ir+ih].lstrip('\t').split(delimiter)
        for ic in range(cols):
            data[ic, ir] = float(tmp[ic])
    print('%d row(s), %d column(s) data loaded'%(rows, cols))
    print('Header:')
    for headerline in header:
        print(headerline)
    print('Endnote:')
    for endline in endnote:
        print(endline)
    if returnnotes:
        return data, header, endnote
    else:
        return data

####    ____HIDDEN METHODS____    ####

def _correct_bias_offset(data, fileType):
    try:
        if fileType == '.dat':
            I = data.I
        elif fileType == '.3ds':
            I = [np.mean(data.I[ix]) for ix, __ in enumerate(data.en)]
        else:
            print('ERR: Bias offset for {:} not yet implemented'.format(fileType))
            return data
        for ix, (I_low, I_high) in enumerate(zip(I[:-1], I[1:])):
            if np.sign(I_low) != np.sign(I_high):
                en_low, en_high = data.en[ix], data.en[ix+1]
                biasOffset = en_high - I_high * (en_high-en_low) / (I_high - I_low)
                data.en -= biasOffset
                break
        print('Corrected for a bias offset of {:2.2f} meV'.format(biasOffset*1000))
        return data
    except:
        print('ERR: File not in standard format for processing. Could not correct for Bias offset')
        return data

def _nice_units(data):
    '''Switch to commonly used units.

    fileType    - .3ds : Use nS for LIY and didv attribute

    History:
        2017-08-10  - HP : Comment: Missing a factor of 2, phase error not
                           justified 
    '''
    def use_nS(data):
        def chi(X):
            gFit = X * data.didv / lockInMod
            err = np.absolute(gFit - didv)
            return np.log(np.sum(err**2))
        lockInMod = float(data.header['Lock-in>Amplitude'])
        current = np.mean(data.I, axis=(1,2))
        didv = np.gradient(current) / np.gradient(data.en)
        result = minimize(chi, 1)
        data.to_nS = result.x / lockInMod * 1e9 
        data.didv *= data.to_nS
        data.LIY *= data.to_nS
        data.didvStd *= data.to_nS
        phi = np.arccos(1.0/result.x)
    
    def use_nm(data):
        fov = [float(val) for val in data.header['Scan>Scanfield'].split(';')]
        data.x = 1e9 * np.linspace(0, fov[2], data.Z.shape[1])
        data.y = 1e9 * np.linspace(0, fov[3], data.Z.shape[0])
        data.qx = stmpy.tools.fftfreq(len(data.x), data.x[-1])
        data.qy = stmpy.tools.fftfreq(len(data.y), data.y[-1])
        data.Z *= 1e9
        data._pxToNm = data.x[-1]/len(data.x)
        data._pxToInvNm = data.qx[-1]/len(data.qx)
        print('WARNING: I am not 100% sure that the q scale is right...')
    
    use_nS(data)
    use_nm(data)
    return data


####    ____CLASS DEFINITIONS____   ####

class Nanonis3ds(object):
    '''Data structure for Nanonis DOS maps.

    Attributes:
        self.Z      - 2D numpy array containing topography channel. Looks for
                      the 'setup Z' recored simultaneously with the DOS map, if
                      not found it will resort to the 'scan Z' measured when
                      moving between lines. 
        self.I      - 3D numpy array for current at each poont.
        self.LIY    - 3D numpy array containing lock-in Y channel.
        self.didv   - 1D numpy array for average LIY
        self.didvStd - 1D numpy array for standard deviation in dIdV.
        self.en     - 1D numpy array for energies used in bias sweep. 
        self.header - Dictionary of all recorded experimental parameters.
        self.grid   - Dictionary of all grid spectroscopy data (individual sweeps, etc.)
        self.scan   - Dictionary with all scan data.

    Methods:
        None

    History:
        2017-08-06  - HP : Added support for recangular maps by changing the
                           order in which data points are read.
        2017-08-10  - HP : Added support for non-linear spaced energies.
    '''
    def __init__(self, filePath):
        if self._load3ds(filePath):
            LIYNames =  ['LIY 1 omega (A)', 'LIY 1 omega [AVG] (A)']
            if self._make_attr('LIY', LIYNames, 'grid'):
                self.didv = np.mean(self.LIY, axis=(1,2))
                self.didvStd = np.std(self.LIY, axis=(1,2))
            else:
                print('ERR: LIY AVG channel not found, resort to manual ' + 
                      'definitions.  Found channels:\n {:}'.format(self.data.keys()))
            
            self._make_attr('I',  ['Current (A)', 'Current [AVG] (A)'], 'grid')
            if self._make_attr('Z',  ['Z (m)', 'Z [AVG] (m)'], 'grid'):
                self.Z = self.Z[0]
            else:
                self._make_attr('Z', ['Scan:Z (m)'], 'scan')
                print('WARNING: Using scan channel for Z attribute.')
            try:     
                self.en = np.mean(self.grid['Bias [AVG] (V)'], axis=(1,2))
            except KeyError:
                print('WARNING: Assuming energy layers are evenly spaced.')
                self.en = np.linspace(self.scan['Sweep Start'].flatten()[0],
                                      self.scan['Sweep End'].flatten()[0],
                                      self._info['points'])
        else: 
            raise NameError('File not found.')

    def _make_attr(self, attr, names, data):
        '''
        Trys to give object an attribute from self.data by looking through
        each key in names.  It will add only the fist match, so the order of
        names dictates the preferences.

        Inputs:
            attr    - Required : Name of new attribute
            names   - Required : List of names to search for
            data    - Required : Name of a current attribute in which the new
                                 attribute is stored.

        Returns:
            1   - If successfully added the attribute
            0   - If name is not found.

        History:
            2017-08-11  - HP : Initial commit.
            2017-08-24  - HP : Now uses grid z value for Z attribute.
        '''
        dat = getattr(self, data)
        for name in names:
            if name in dat.keys():
                setattr(self, attr, dat[name])
                return 1
        return 0

    def _load3ds(self, filePath):
        try: 
            fileObj = open(filePath, 'rb')
        except: 
            return 0
        self.header={}
        while True:
            line = fileObj.readline().strip().decode('utf-8')
            if line == ':HEADER_END:': 
                break
            splitLine = line.split('=')
            self.header[splitLine[0]] = splitLine[1]

        self._info = {'params'	: int(self.header['# Parameters (4 byte)']),
                    'paramName'	: self.header['Fixed parameters'][1:-1].split(';') +
                                  self.header['Experiment parameters'][1:-1].split(';'),
                    'channels'	: self.header['Channels'][1:-1].split(';'),
                    'points'	: int(self.header['Points']),
                    'sizex' 	: int(self.header['Grid dim'][1:-1].split(' x ')[0]),
                    'sizey'	: int(self.header['Grid dim'][1:-1].split(' x ')[1]),
                    'dataStart'	: fileObj.tell()
                     }

        self.grid = {}; self.scan = {}
        for channel in self._info['channels']:
            self.grid[channel] = np.zeros(
                    [self._info['points'], self._info['sizey'], self._info['sizex']])
        for channel in self._info['paramName']:
            self.scan[channel] = np.zeros([self._info['sizey'], self._info['sizex']])

        try:
            for iy in range(self._info['sizey']):
                for ix in range(self._info['sizex']):
                    for channel in self._info['paramName']:
                        value = unpack('>f',fileObj.read(4))[0]
                        self.scan[channel][iy,ix] = value

                    for channel in self._info['channels']:
                        for ie in range(self._info['points']):
                            value = unpack('>f',fileObj.read(4))[0]
                            self.grid[channel][ie,iy,ix] = value
        except:
            print('WARNING: Data set is not complete.')

        dataRead = fileObj.tell()
        fileObj.read()
        allData = fileObj.tell()
        if dataRead == allData: 
            print('File import successful.')
        else: 
            print('ERR: Did not reach end of file.')
        fileObj.close()
        return 1

class LongTermSpectrum(object):
    '''
header: a dict containging all parameters
time: time length of the spectrum, in unit of (s)
freq: freq range of the spectrum, in unit of (Hz)
fftI, fftV, or fftSignal:
    current signal >> fftI
    voltage signal >> fftV
    other signal   >> fftSignal 
Example Usage:
import stmpy \nimport matplotlib.pyplot as plt \nimport matplotlib.dates as md \nimport numpy as np \nfrom datetime import datetime \n 
data = stmpy.load('***.nsp')
x=np.array(data.start,dtype='datetime64[s]') \nstepsize = int(np.floor(((data.end - data.start).total_seconds())/data.header['DATASIZEROWS']))
dates = x + np.arange(0,stepsize*data.header['DATASIZEROWS'],stepsize) \nnew_dates=[np.datetime64(ts).astype(datetime) for ts in dates]
datenums=md.date2num(new_dates) \nplt.subplots_adjust(bottom=0.2) \nplt.xticks( rotation=45 ) \nplt.ax=plt.gca() \nxfmt = md.DateFormatter('%Y-%m-%d %H:%M:%S')
plt.ax.xaxis.set_major_formatter(xfmt) \nplt.pcolormesh(datenums, data.freq, data.fftI)#depending on which kind of data used
plt.clim(0,1e-12)#can be different \nplt.ylabel('Frequency (Hz)') \nplt.savefig('pic name.png',dpi = 600,bbox_inches = 'tight') \nplt.show()
    '''
    def __init__(self, filePath):
        self._loadnsp(filePath)
        if self.header['SIGNAL'] == 'Current (A)':
            self.fftI = self.data.T
        elif self.header['SIGNAL'] == 'InternalGeophone (V)':
            self.fftV = self.data.T
        else:
            self.fftSignal = self.data.T

    def _loadnsp(self, filePath):
        try: fileObj = open(filePath, 'rb')
        except: return 0
        self.header = {}
        while True:
            line = fileObj.readline().strip().decode('utf-8')
            if line == ':HEADER_END:': 
                break
            elif re.match('^:.*:$', line):
                tagname = line[1:-1]
            else:
                try:
                    self.header[tagname] = int(line.split('\t')[0])
                except:
                    self.header[tagname] = line.split('\t')[0]

        self.freq = np.linspace(0, np.round(float(self.header['DATASIZECOLS'])*float(self.header['DELTA_f'])),float(self.header['DATASIZECOLS']))
        
        self.start = datetime.strptime(self.header['START_DATE']+self.header['START_TIME'],'%d.%m.%Y%H:%M:%S')
        self.end = datetime.strptime(self.header['END_DATE']+self.header['END_TIME'],'%d.%m.%Y%H:%M:%S')
        self.time = np.linspace(0, (self.end - self.start).total_seconds(), int(self.header['DATASIZEROWS']))

        self.data = np.zeros([int(self.header['DATASIZEROWS']),int(self.header['DATASIZECOLS'])])
        fileObj.read(2) #first two bytes are not data
        try:
            for ix in range(int(self.header['DATASIZEROWS'])):
                for iy in range(int(self.header['DATASIZECOLS'])):
                    value = unpack('>f',fileObj.read(4))[0]
                    self.data[ix,iy] = value
        except:
            print('Error: Data set is not complete')


class NanonisSXM(object):
    def __init__(self, filename):
        self.header = {}
        self.header['filename'] = filename
        self._open()
    
    def _open(self):
        self._file = open(os.path.normpath(self.header['filename']), 'rb')
        s1 = self._file.readline().decode('utf-8')
        if not re.match(':NANONIS_VERSION:', s1):
            print('The file %s does not have the Nanonis SXM'.format(self.header['filename']))
            return
        self.header['version'] = int(self._file.readline())
        while True:
            line = self._file.readline().strip().decode('utf-8')
            if re.match('^:.*:$', line):
                tagname = line[1:-1]
            else:
                if 'Z-CONTROLLER' == tagname:
                    keys = line.split('\t')
                    values = self._file.readline().strip().decode('utf-8').split('\t')
                    self.header['z-controller'] = dict(zip(keys, values))
                elif tagname in ('BIAS', 'REC_TEMP', 'ACQ_TIME', 'SCAN_ANGLE'):
                    self.header[tagname.lower()] = float(line)
                elif tagname in ('SCAN_PIXELS', 'SCAN_TIME', 'SCAN_RANGE', 'SCAN_OFFSET'):
                    self.header[tagname.lower()] = [ float(i) for i in re.split('\s+', line) ]
                elif 'DATA_INFO' == tagname:
                    if 1 == self.header['version']:
                        keys = re.split('\s\s+',line)
                    else:
                        keys = line.split('\t')
                    self.header['data_info'] = []
                    while True:
                        line = self._file.readline().strip().decode('utf-8')
                        if not line:
                            break
                        values = line.strip().split('\t')
                        self.header['data_info'].append(dict(zip(keys, values)))
                elif tagname in ('SCANIT_TYPE','REC_DATE', 'REC_TIME', 'SCAN_FILE', 'SCAN_DIR'):
                    self.header[tagname.lower()] = line
                elif 'SCANIT_END' == tagname:
                    break
                else:
                    if tagname.lower() not in self.header:
                        self.header[tagname.lower()] = line
                    else:
                        self.header[tagname.lower()] += '\n' + line
        if 1 == self.header['version']:
            self.header['scan_pixels'].reverse()
        self._file.readline()
        self._file.read(2) # Need to read the byte \x1A\x04, before reading data
        size = int( self.header['scan_pixels'][0] * self.header['scan_pixels'][1] * 4)
        shape = [int(val) for val in self.header['scan_pixels']]
        self.channels = {}
        for channel in self.header['data_info']:
            if channel['Direction'] == 'both':
                self.channels[channel['Name'] + '_Fwd'] = np.ndarray(
                        shape=shape[::-1], dtype='>f', buffer=self._file.read(size))
                self.channels[channel['Name'] + '_Bkd'] = np.ndarray(
                        shape=shape[::-1], dtype='>f', buffer=self._file.read(size))
            else:
                self.channels[channel['Name'] + channel['Direction']] = np.ndarray(shape=shape, dtype='>f', buffer=self._file.read(size))
        try:
            self.Z = self.channels['Z_Fwd']
            self.I = self.channels['Current_Fwd']
            self.LIY = self.channels['LIY_1_omega_Fwd']
        except KeyError: print('WARNING:  Could not create standard attributes, look in channels instead.')
        self._file.close()


class NanonisDat(object):
    def __init__(self,filePath):
        if self._loadDat(filePath):
            self._make_attr('didv', 
                    ['LIY 1 omega (A)', 'LIY 1 omega [AVG] (A)'], 'channels')
            self._make_attr('I', ['Current (A)', 'Current [AVG] (A)'],
            'channels')
            self._make_attr('en', ['Bias (V)', 'Bias calc (V)'], 'channels')
            if 'LIY 1 omega [00001] (A)' in self.channels.keys():
                sweeps = int(self.header['Bias Spectroscopy>Number of sweeps'])
                self.LIY = np.zeros([len(self.en), sweeps])
                for ix in range(1, sweeps+1):
                    s = str(ix).zfill(5)
                    self.LIY[:,ix-1] = self.channels['LIY 1 omega [' + s + '] (A)']
                self.didvStd = np.std(self.LIY, axis=1)

    def _make_attr(self, attr, names, data):
        '''
        Trys to give object an attribute from self.data by looking through
        each key in names.  It will add only the fist match, so the order of
        names dictates the preferences.

        Inputs:
            attr    - Required : Name of new attribute
            names   - Required : List of names to search for
            data    - Required : Name of a current attribute in which the new
                                 attribute is stored.

        Returns:
            1   - If successfully added the attribute
            0   - If name is not found.

        History:
            2017-08-29  - HP : Initial commit (copied from Nanonis3ds)
        '''
        dat = getattr(self, data)
        for name in names:
            if name in dat.keys():
                setattr(self, attr, dat[name])
                return 1
        return 0

    def _loadDat(self, filePath):
        self.channels = {}
        self.header = {}
        fileObj = open(filePath,'r')
        while True:
            line = fileObj.readline()
            splitLine = line.split('\t')
            if line[0:6] == '[DATA]': 
                break
            elif line.rstrip() != '': 
                self.header[splitLine[0]] = splitLine[1]
        channels = fileObj.readline().rstrip().split('\t')
        allData = []
        for line in fileObj:
            line = line.rstrip().split('\t')
            allData.append(np.array(line, dtype = float))
        allData = np.array(allData)
        for ix, channel in enumerate(channels):
            self.channels[channel] = allData[:,ix]
        dataRead = fileObj.tell()
        fileObj.read()
        finalRead = fileObj.tell()
        if dataRead == finalRead: 
            print('File import successful.')
            fileObj.close()
            return 1
        else: 
            print('ERR: Did not reach end of file.')
            fileObj.close()
            return 0

        
        

class NISTnvi(object):
    def __init__(self,nviData):
        self._raw = nviData['imagetosave']
        self.map = self._raw.currentdata[0]
        self.header = {name:self._raw.header[0][name][0] for name in self._raw.header[0].dtype.names}
        self.info = {'FILENAME'    : self._raw.filename[0],
                     'FILSIZE'     : int(self._raw.header[0].filesize[0]),
                     'CHANNELS'    : self._raw.header[0].scan_channels[0],
                     'XSIZE'       : self._raw.xsize[0],
                     'YSIZE'       : self._raw.ysize[0],
                     'TEMPERATURE' : self._raw.header[0].temperature[0],
                     'LOCKIN_AMPLITUDE' : self._raw.header[0].lockin_amplitude[0],
                     'LOCKIN_FREQUENCY' : self._raw.header[0].lockin_frequency[0],
                     'DATE'        : self._raw.header[0].date[0],
                     'TIME'        : self._raw.header[0].time[0],
                     'BIAS_SETPOINT'    : self._raw.header[0].bias_setpoint[0],
                     'BIAS_OFFSET' : self._raw.header[0].bias_offset[0],
                     'BFIELD'      : self._raw.header[0].bfield[0],
                     'ZUNITS'      : self._raw.zunits[0],
					}
        
class NISTnvl(object):
    def __init__(self,nvlData):
        self._raw = nvlData['savestructure']
        self.en = self._raw.energies[0]
        self.map = self._raw.fwddata[0]
        self.ave = [np.mean(layer) for layer in self.map]
        self.header = {name:self._raw.header[0][name][0] for name in self._raw.header[0].dtype.names}
        for name in self._raw.dtype.names:
            if name not in self.header.keys():
                self.header[name] = self._raw[name][0]
        self.info = {}
        try:
            self.info['FILENAME']   = self._raw.filename[0]
        except:
            1
        try:
            self.info['FILSIZE']    = int(self._raw.header[0].filesize[0])
        except:
            1
        try:
            self.info['CHANNELS']   = self._raw.header[0].scan_channels[0]
        except:
            1
        try:
            self.info['XSIZE']      = self._raw.xsize[0]
        except:
            1
        try:
            self.info['YSIZE']      = self._raw.ysize[0]
        except:
            1
        try:
            self.info['TEMPERATURE']= self._raw.header[0].temperature[0]
        except:
            1
        try:
            self.info['LOCKIN_AMPLITUDE']= self._raw.header[0].lockin_amplitude[0]
        except:
            1
        try:
            self.info['LOCKIN_FREQUENCY']= self._raw.header[0].lockin_frequency[0]
        except:
            1
        try:
            self.info['DATE']       = self._raw.header[0].date[0]
        except:
            1
        try:
            self.info['TIME']       = self._raw.header[0].time[0]
        except:
            1
        try:
            self.info['BIAS_SETPOINT'] = self._raw.header[0].bias_setpoint[0]
        except:
            1
        try:
            self.info['BIAS_OFFSET']= self._raw.header[0].bias_offset[0]
        except:
            1
        try:
            self.info['BFIELD']     = self._raw.header[0].bfield[0]
        except:
            1
        try:
            self.info['WINDOWTITLE'] = self._raw.windowtitle[0]
        except:
            1
        try:
            self.info['XYUNITS']    = self._raw.xyunits[0]
        except:
            1
        try:
            self.info['EUNITS']     = self._raw.eunits[0]
        except:
            1

class AsciiFile(object):
    def __init__(self, filePath):
        self.load(filePath)
    
    def load(self, filePath):
        fid = open(filePath, 'r')
        header= {}
        channels = {}
        while True:
            line = fid.readline().rstrip()
            if line is '':
                break
            splitLine = line.split(':')
            header[splitLine[0]] = splitLine[1]
        channelNames = fid.readline().rstrip().split('      ')
        for chn in channelNames:
            channels[chn] = []
        for data in fid.readlines():
            dsplit = data.rstrip().split('   ')
            dfloat = [float(val) for val in dsplit]
            for chn, val in zip(channelNames, dfloat):
                channels[chn] += [val]
        for chn in channelNames:
            channels[chn] = np.array(channels[chn])
        if len(channelNames) is 2:
            self.x = channels[channelNames[0]]
            self.y = channels[channelNames[1]]
        self.header = header
        self.channels = channels
        fid.close()





from __future__ import print_function
try:
    import cv2
except ModuleNotFoundError:
    print("Please install opencv-python module using following command:\npip3 install opencv-python")
import stmpy
import numpy as np
import scipy as sp
import matplotlib as mpl
import matplotlib.pyplot as plt
import scipy.optimize as opt
import scipy.ndimage as snd
from scipy.interpolate import interp1d, interp2d
from skimage import transform as tf
from skimage.feature import peak_local_max
#import stmpy.driftcorr as dfc

'''
Local drift correction of square/triangular lattice. (Please carefully rewrite if a rectangular lattice version is needed.)

Usage:
0. please import stmpy.driftcorr (Package 'skimage' required, run 'pip install -U scikit-image' in terminal);

1. findBraggs: FT the topo image (please interpolate to be 2D square array if not), then find all Bragg peaks by peak_local_max, plot result with abs(FT) to check validity (sometimes points in the center should be removed);

2. gshearcorr: global shear correction, outputs corrected image and positions of the corrected Bragg peaks;

3. phasemap: use the new Bragg peaks to generate local phase shift (theta) maps. Use chkphasemap get phase (phi) maps and check if pi lines in those matches with raw image. If not, please check your Bragg peaks or adjust sigma in the FT DC filter;

4. fixphaseslip: fix phase slips of 2d phase maps in a 1d spiral way.(this is not well developed yet. Phase slips could still be present
near edges after processing, please crop images and DOS maps accordingly and manually AFTER applying driftmap.) Output is 2d array in the same shape as the input array;

5. driftmap: calculate drift fields u(x,y) in both x and y directions;

6. driftcorr: local drift correction using u(x,y) on 2D (topo) or 3D(DOS map) input.

7. (OPTIONAL) chkperflat: check perfect lattice generated by Q1, Q2 (to be improved)

REFERENCES:
[1] MH Hamidian, et al. "Picometer registration of zinc impurity states in Bi2Sr2CaCu2O8+d for phase determination in intra-unit-cell Fourier transform STM", New J. Phys. 14, 053017 (2012).
[2] JA Slezak, PhD thesis (Ch. 3), http://davisgroup.lassp.cornell.edu/theses/Thesis_JamesSlezak.pdf

History:
    2017-04-28      CREATED BY JIANFENG GE
    04/29/2019      RL : Add documents for all functions. Add another method to calculate phasemap.
                            Add inverse FFT method to apply the drift field.
'''

#1. - getAttrs
def getAttrs(obj, a0, size=None, pixels=None):
    '''
    Create attributes of lattice constant, map size, number of pixels, and qscale for Spy object.

    Input:
        obj         - Required : Spy object of topo (2D) or map (3D).
        a0             - Required : Lattice constant in the unit of nm.
        size         - Optional : Size of the map in the unit of nm. If not offered, it'll be created
                                    automatically from header file.
        pixels         - Optional : Number of pixels of the topo/map. If not offered, it'll be created
                                    automatically from header file.

    Returns:
        N/A

    Usage:
        import stmpy.driftcorr as dfc
        dfc.getAttrs(topo, a0=a0)

    '''
    if size is None:
        try:
            size = obj.header['scan_range'][-1]
        except KeyError:
            try:
                size = float(obj.header['Grid settings'].split(";")[-2])
            except:
                print("Error: Cannot find map size from header. Please input it manually.")
    if pixels is None:
        try:
            pixels = int(obj.header['scan_pixels'][-1])
        except KeyError:
            try:
                pixels = int(obj.header['Grid dim'].split()[-1][:-1])
            except:
                print("Error: Cannot find number of pixels from header. Please input it manually.")
    obj.a0 = a0
    obj.size = size * 1e9
    obj.pixels = pixels
    obj.qmag = obj.size / obj.a0
    obj.qscale = obj.pixels / (2*obj.qmag)

#2. - findBraggs
def findBraggs(A, obj=None, rspace=True, min_dist=5, thres=0.25, r=0.25, \
    w=None, maskon=True, show=False, angle=0, even_out=True, update_obj=True):
    '''
    Find Bragg peaks in the unit of pixels of topo or FT pattern A using peak_local_max. If obj is offered,
    an attribute of bp will be created for obj.

    Input:
        A           - Required : 2D array of topo in real space, or FFT in q space.
        obj         - Optional : Object associated with A
        min_dist    - Optional : Minimum distance (in pixels) between peaks. Default: 5
        thres       - Optional : Minimum intensity of Bragg peaks relative to max value. Default: 0.25
        rspace      - Optional : Boolean indicating if A is real or Fourier space image. Default: True
        r           - Optional : width of the gaussian mask to remove low-q noise, =r*width
        w           - Optional : width of the mask that filters out noise along qx=0 and qy=0 lines.
                                    Set w=None will disable this mask.
        angle       - Optional : Angle of line masks in degrees.
        maskon      - Optional : Boolean, if False then no mask will be applied.
        show        - Optional : Boolean, if True then A and Bragg peaks will be plotted out.
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.

    Returns:
        coords      -  (4x2) array contains Bragg peaks in the format of [[x1,y1],[x2,y2],...,[x4,y4]]

    Usage:
        import stmpy.driftcorr as dfc
        bp = dfc.findBraggs(A, obj=topo, min_dist=10, thres=0.2, rspace=True, show=True)

    History:
        04/28/2017      JG : Initial commit.
        04/29/2019      RL : Add maskon option, add outAll option, and add documents.

    '''
    if rspace is True:
        F = stmpy.tools.fft(A, zeroDC=True)
    else:
        F = np.copy(A)
    # Remove low-q high intensity data with Gaussian mask
    if maskon is True:
        *_, Y, X = np.shape(A)
        Lx = X * r
        Ly = Y * r
        x = np.arange(X)
        y = np.arange(Y)
        p0 = [int(X/2), int(Y/2), Lx, Ly, 1, np.pi/2]
        G = 1-stmpy.tools.gauss2d(x, y, p=p0)
        if w is not None:
            mask3 = np.ones([Y, X])
            mask3[Y//2-int(Y*w):Y//2+int(Y*w),:] = 0
            mask3[:,X//2-int(X*w):X//2+int(X*w)] = 0
        else:
            mask3 = 1
        F *= G * mask3
        if obj is not None:
            L = np.shape(A)[-1]
            x = np.arange(L)
            y = np.arange(L)
            mask2 = stmpy.tools.gauss_ring(x, y, major=obj.qmag, minor=obj.qmag,
                                   sigma=10, x0=L/2, y0=L/2)
            F *= mask2*mask3
    coords = peak_local_max(F, min_distance=min_dist, threshold_rel=thres)
    coords = np.fliplr(coords)

    if show is True:
        plt.figure(figsize=[4,4])
        c = np.mean(F)
        s = np.std(F)
        plt.imshow(F, cmap=plt.cm.gray_r, interpolation='None', origin='lower left', clim=[0,c+5*s], aspect=1)
        plt.plot(coords[:, 0], coords[:, 1], 'r.')
        plt.gca().set_aspect(1)
        plt.axis('tight')
        print('#:\t[x y]')
        for ix, iy in enumerate(coords):
            print(ix, end='\t')
            print(iy)
    if even_out is True:
        center = np.array(np.shape(A)[::-1]) // 2
        for i, ix in enumerate(coords):
            coords[i] = center + (ix - center) // 2 * 2
    if obj is not None:
        if update_obj is True:
            obj.bp = coords
    return coords

#3. - gshearcorr
def gshearcorr(A, bp=None, obj=None, rspace=True, pts1=None, pts2=None, angle=np.pi/4, matrix=None, update_obj=True):
    '''
    Global shear correction based on position of Bragg peaks in FT of 2D or 3D array A

    Inputs:
        A           - Required : 2D or 3D array to be shear corrected.
        bp          - Required : (Nx2) array contains Bragg peaks in the unit of pixels.
        obj         - Optional : Spy object of topo (2D) or map (3D).
        rspace      - Optional : Boolean indicating if A is real or Fourier space image. Default: True
        pts1        - Optional : 3x2 array containing coordinates of three points in the raw FT (center,
                                    bg_x, bg_y).
        pts2        - Optional : 3x2 array containing coordinates of three corresponding points in the corrected
                                    FT (i.e., model center and bg_x and bg_y coordinates).
        angle       - Optional : Specify angle between scan direction and lattice unit vector direction (x and ux direction)
                                    in the unit of radian. Default is pi/4 -- 45 degrees rotated.
        matrix      - Optional : If provided, matrix will be used to transform the dataset directly
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.

    Returns:
        A_corr      - 2D or 3D array after global shear correction.
        M           - Transformation matrix to shear correct the topo/map

    Usage:
        import stmpy.driftcorr as dfc
        M, A_gcorr = dfc.gshearcorr(A, bp, obj=topo, rspace=True)
    '''
    *_, s2, s1 = np.shape(A)
    bp_temp = bp
    if matrix is None:
        if pts1 is None:
            bp = sortBraggs(bp, s=np.shape(A))
            s = np.array(np.shape(A))
            bp_temp = bp * s
            center = [int(s[0]*s[1]/2), int(s[0]*s[1]/2)]
            Q1, Q2, Q3, Q4, *_ = bp_temp
            if obj is None:
                Qx_mag = compute_dist(Q1, center)
                Qy_mag = compute_dist(Q2, center)
                Q_corr = np.mean([Qx_mag, Qy_mag])
            else:
                Q_corr = obj.qmag
            Qc1 = Q_corr*np.array([-np.cos(angle), -np.sin(angle)]) + center
            Qc2 = Q_corr*np.array([np.sin(angle), -np.cos(angle)]) + center
            Q1, Q2, Q3, Q4, *_ = bp
            Qc2 = Qc2 / s
            Qc1 = Qc1 / s
            center = [int(s2/2),int(s1/2)]
            print(Q1,Q2,Qc1,Qc2,center)
            pts1 = np.float32([center,Q1,Q2])
        else:
            pts1 = pts1.astype(np.float32)
        if pts2 is None:
            pts2 = np.float32([center,Qc1,Qc2])
        else:
            pts2 = pts2.astype(np.float32)
        M = cv2.getAffineTransform(pts1,pts2)
    else:
        M = matrix

    if rspace is not True:
        A_corr = cv2.warpAffine(A, M, (s2,s1),
                        flags=(cv2.INTER_CUBIC + cv2.BORDER_CONSTANT))
    else:
        M[:,-1] = np.array([0,0])
        offset = np.min(A)
        A = A - offset
        A_corr = cv2.warpAffine(np.flipud(A.T), M, (s2,s1),
                        flags=(cv2.INTER_CUBIC + cv2.BORDER_CONSTANT))
        A_corr = np.flipud(A_corr).T + offset
    return M, A_corr


#4. phasemap
def phasemap(A, bp, obj=None, sigma=10, method="lockin", update_obj=True):
    '''
    Calculate local phase and phase shift maps. Two methods are available now: spatial lockin or Gaussian mask convolution

    Input:
        A           - Required : 2D arrays after global shear correction with bad pixels cropped on the edge
        bp          - Required : Coords of Bragg peaks of FT(A), can be computed by findBraggs(A)
        obj         - Optional : Spy object of topo (2D) or map (3D).
        sigma       - Optional : width of DC filter in lockin method or len(A)/s
        method      - Optional : Specify which method to use to calculate phase map.
                                "lockin": Spatial lock-in method to find phase map
                                "convolution": Gaussian mask convolution method to find phase map
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.

    Returns:
        thetax      -       2D array, Phase shift map in x direction, relative to perfectly generated cos lattice
        thetay      -       2D array, Phase shift map in y direction, relative to perfectly generated cos lattice
        Q1          -       Coordinates of 1st Bragg peak
        Q2          -       Coordinates of 2nd Bragg peak

    Usage:
        import stmpy.driftcorr as dfc
        thetax, thetay, Q1, Q2 = dfc.phasemap(A, bp, sigma=10, method='lockin')

    History:
        04/28/2017      JG : Initial commit.
        04/29/2019      RL : Add "convolution" method, and add documents.
        11/30/2019      RL : Add support for non-square dataset
    '''

    *_, s2, s1 = A.shape
    s = np.minimum(s1, s2)
    bp = sortBraggs(bp, s=np.shape(A))
    t1 = np.arange(s1, dtype='float')
    t2 = np.arange(s2, dtype='float')
    x, y = np.meshgrid(t1, t2)
    Q1 = 2*np.pi*np.array([(bp[0][0]-int(s1/2))/s1, (bp[0][1]-int(s2/2))/s2])
    Q2 = 2*np.pi*np.array([(bp[1][0]-int(s1/2))/s1, (bp[1][1]-int(s2/2))/s2])
    if method is "lockin":
        Axx = A * np.sin(Q1[0]*x+Q1[1]*y)
        Axy = A * np.cos(Q1[0]*x+Q1[1]*y)
        Ayx = A * np.sin(Q2[0]*x+Q2[1]*y)
        Ayy = A * np.cos(Q2[0]*x+Q2[1]*y)
        Axxf = FTDCfilter(Axx, sigma)
        Axyf = FTDCfilter(Axy, sigma)
        Ayxf = FTDCfilter(Ayx, sigma)
        Ayyf = FTDCfilter(Ayy, sigma)
        thetax = np.arctan2(Axxf, Axyf)
        thetay = np.arctan2(Ayxf, Ayyf)
        if obj is not None:
            if update_obj is True:
                obj.phix = thetax
                obj.phiy = thetay
                obj.Q1 = Q1
                obj.Q2 = Q2
        return thetax, thetay, Q1, Q2
    elif method is "convolution":
        t_x = np.arange(s1)
        t_y = np.arange(s2)
        xcoords, ycoords = np.meshgrid(t_x, t_y)
        exponent_x = (Q1[0] * xcoords + Q1[1] * ycoords)#(2.* np.pi/s)*(Q1[0] * xcoords + Q1[1] * ycoords)
        exponent_y = (Q2[0] * xcoords + Q2[1] * ycoords)#(2.* np.pi/s)*(Q2[0] * xcoords + Q2[1] * ycoords)
        A_x = A * np.exp(np.complex(0,-1)*exponent_x)
        A_y = A * np.exp(np.complex(0,-1)*exponent_y)
        sx = sigma
        sy = sigma * s1 / s2
        Amp = 1/(4*np.pi*sx*sy)
        p0 = [int(s/2), int(s/2), sx, sy, Amp, np.pi/2]
        G = stmpy.tools.gauss2d(t_x, t_y, p=p0, symmetric=True)
        T_x = sp.signal.fftconvolve(A_x, G, mode='same',)
        T_y = sp.signal.fftconvolve(A_y, G, mode='same',)
        R_x = np.abs(T_x)
        R_y = np.abs(T_y)
        phi_y = np.angle(T_y)
        phi_x = np.angle(T_x)
        if obj is not None:
            if update_obj is True:
                obj.phix = phi_x
                obj.phiy = phi_y
                obj.Q1 = Q1
                obj.Q2 = Q2
        return phi_x, phi_y, Q1, Q2
    else:
        print('Only two methods are available now:\n1. lockin\n2. convolution')

#5. fixphaseslip
def fixphaseslip(A, thres=None, maxval=None, method='unwrap', orient=0):
    '''
    Fix phase slip by adding 2*pi at phase jump lines.

    Inputs:
        A       - Required : 2D arrays of phase shift map, potentially containing phase slips
        thres   - Optional : Float number, specifying threshold for finding phase jumps in diff(A). Default: None
        method  - Optional : Specifying which method to fix phase slips.
                                "unwrap": fix phase jumps line by line in x direction and y direction, respectively
                                "spiral": fix phase slip in phase shift maps by flattening A into a 1D array in a spiral way
        orient  - Optional : Used in "spiral" phase fixing method. 0 for clockwise and 1 for counter-clockwise

    Returns:

        phase_corr      -       2D arrays of phase shift map with phase slips corrected

    Usage:
        import stmpy.driftcorr as dfc
        thetaxf = dfc.fixphaseslip(thetax, method='unwrap')

    History:
        04/28/2017      JG : Initial commit.
        04/29/2019      RL : Add "unwrap" method, and add documents.
    '''
    output = np.copy(A[::-1,::-1])
    if len(np.shape(A)) == 2:
        *_, s2, s1 = np.shape(A)
        for i in range(s2):
            output[i,:] = unwrap_phase(output[i,:], tolerance=thres, maxval=maxval)
        for i in range(s1):
            output[:,i] = unwrap_phase(output[:,i], tolerance=thres, maxval=maxval)
        return output[::-1,::-1]

def unwrap_phase(ph, tolerance=None, maxval=None):
    maxval = 2 * np.pi if maxval is None else maxval0
    tol = 0.25*maxval if tolerance is None else tolerance*maxval
    if len(ph) < 2:
        return ph

    dph = np.diff(ph)
    dph[np.where(np.abs(dph) < tol)] = 0
    dph[np.where(dph < -tol)] = 1
    dph[np.where(dph > tol)] = -1
    ph[1:] += maxval * np.cumsum(dph)
    return ph

#6. driftmap
def driftmap(phix=None, phiy=None, Q1=None, Q2=None, obj=None, method="lockin", update_obj=True):
    '''
    Calculate drift fields based on phase shift maps, with Q1 and Q2 generated by phasemap.

    Inputs:
        obj         - Optional : Spy object of topo (2D) or map (3D).
        phix        - Optional : 2D arrays of phase shift map in x direction with phase slips corrected
        phiy        - Optional : 2D arrays of phase shift map in y direction with phase slips corrected
        Q1          - Optional : Coordinates of 1st Bragg peak, generated by phasemap
        Q2          - Optional : Coordinates of 2nd Bragg peak, generated by phasemap
        method      - Optional : Specifying which method to use.
                                    "lockin": Used for phase shift map generated by lockin method
                                    "convolution": Used for phase shift map generated by lockin method
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.

    Returns:
        ux          - 2D array of drift field in x direction
        uy          - 2D array of drift field in y direction

    Usage:
        import stmpy.driftcorr as dfc
        ux, uy = dfc.driftmap(thetaxf, thetayf, Q1, Q2, method='lockin')

    History:
        04/28/2017      JG : Initial commit.
        04/29/2019      RL : Add "lockin" method, and add documents.
        11/30/2019      RL : Add support for non-square dataset
    '''
    if method is "lockin":
        tx = np.copy(phix)
        ty = np.copy(phiy)
        ux = -(Q2[1]*tx - Q1[1]*ty) / (Q1[0]*Q2[1]-Q1[1]*Q2[0])
        uy = -(Q2[0]*tx - Q1[0]*ty) / (Q1[1]*Q2[0]-Q1[0]*Q2[1])
        if obj is not None:
            if update_obj is True:
                obj.ux = ux
                obj.uy = uy
        return ux, uy
    elif method is "convolution":
        #s = np.shape(thetax)[-1]
        Qx_mag = np.sqrt((Q1[0])**2 + (Q1[1])**2)
        Qy_mag = np.sqrt((Q2[0])**2 + (Q2[1])**2)
        Qx_ang = np.arctan2(Q1[1],Q1[0]) # in radians
        Qy_ang = np.arctan2(Q2[1],Q2[0]) # in radians
        Qxdrift = 1/(Qx_mag) * phix#s/(2*np.pi*Qx_mag) * thetax
        Qydrift = 1/(Qy_mag) * phiy#s/(2*np.pi*Qy_mag) * thetay
        ux = Qxdrift * np.cos(Qx_ang) - Qydrift * np.sin(Qy_ang-np.pi/2)
        uy = Qxdrift * np.sin(Qx_ang) + Qydrift * np.cos(Qy_ang-np.pi/2)
        if obj is not None:
            if update_obj is True:
                obj.ux = -ux
                obj.uy = -uy
        return -ux, -uy
    else:
        print("Only two methods are available now:\n1. lockin\n2. convolution")

#7. - driftcorr
def driftcorr(A, ux=None, uy=None, obj=None, method="lockin", interpolation='cubic'):
    '''
    Correct the drift in the topo according to drift fields

    Inputs:
        A           - Required : 2D or 3D arrays of topo to be drift corrected
        obj         - Optional : Spy object of topo (2D) or map (3D).
        ux          - Optional : 2D arrays of drift field in x direction, generated by driftmap()
        uy          - Optional : 2D arrays of drift field in y direction, generated by driftmap()
        method      - Optional : Specifying which method to use.
                                    "interpolate": Interpolate A and then apply it to a new set of coordinates,
                                                    (x-ux, y-uy)
                                    "convolution": Used inversion fft to apply the drift fields
        interpolation - Optional : Specifying which method to use for interpolating

    Returns:
        A_corr      - 2D or 3D array of topo with drift corrected

    Usage:
        import stmpy.driftcorr as dfc
        A_corr = dfc.driftcorr(ux, uy, method='interpolate', interpolation='cubic')

    History:
        04/28/2017      JG : Initial commit.
        04/29/2019      RL : Add "invfft" method, and add documents.
        11/30/2019      RL : Add support for non-square dataset
    '''
    if method is "lockin":
        A_corr = np.zeros_like(A)
        *_, s2, s1 = np.shape(A)
        t1 = np.arange(s1, dtype='float')
        t2 = np.arange(s2, dtype='float')
        x, y = np.meshgrid(t1, t2)
        xnew = (x - ux).ravel()
        ynew = (y - uy).ravel()
        tmp = np.zeros(s1*s2)
        if len(A.shape) is 2:
            tmp_f = interp2d(t1, t2, A, kind=interpolation)
            for ix in range(tmp.size):
                tmp[ix] = tmp_f(xnew[ix], ynew[ix])
            A_corr = tmp.reshape(s2, s1)
            return A_corr
        elif len(A.shape) is 3:
            for iz, layer in enumerate(A):
                tmp_f = interp2d(t1, t2, layer, kind=interpolation)
                for ix in range(tmp.size):
                    tmp[ix] = tmp_f(xnew[ix], ynew[ix])
                A_corr[iz] = tmp.reshape(s2, s1)
                print('Processing slice %d/%d...'%(iz+1, A.shape[0]), end='\r')
            return A_corr
        else:
            print('ERR: Input must be 2D or 3D numpy array!')
    elif method is "convolution":
        A_corr = np.zeros_like(A)
        if len(A.shape) is 2:
            return _apply_drift_field(A, ux=ux, uy=uy, zeroOut=True)
        elif len(A.shape) is 3:
            for iz, layer in enumerate(A):
                A_corr[iz] = _apply_drift_field(layer, ux=ux, uy=uy, zeroOut=True)
                print('Processing slice %d/%d...'%(iz+1, A.shape[0]), end='\r')
            return A_corr
        else:
            print('ERR: Input must be 2D or 3D numpy array!')

def _apply_drift_field(A, ux, uy, zeroOut=True):
    A_corr = np.copy(A)
    *_, s2, s1 = np.shape(A)
    t1 = np.arange(s1, dtype='float')
    t2 = np.arange(s2, dtype='float')
    x, y = np.meshgrid(t1, t2)
    xshifted = x - ux
    yshifted = y - uy
    if zeroOut is True:
        A_corr[np.where(xshifted < 0)] = 0
        A_corr[np.where(yshifted < 0)] = 0
        A_corr[np.where(xshifted > s1)] = 0
        A_corr[np.where(yshifted > s2)] = 0
    qcoordx = (2*np.pi/s1)*(np.arange(s1)-int(s1/2))
    qcoordy = (2*np.pi/s2)*(np.arange(s2)-int(s2/2))
    #qcoord = (2*np.pi/s)*(np.arange(s)-(s/2))
    xshifted = np.reshape(xshifted, [1, s1*s2])
    yshifted = np.reshape(yshifted, [1, s1*s2])
    qcoordx = np.reshape(qcoordx, [s1, 1])
    qcoordy = np.reshape(qcoordy, [s2, 1])
    xphase = np.exp(-1j*(np.matmul(xshifted.T, qcoordx.T).T))
    yphase = np.exp(-1j*(np.matmul(yshifted.T, qcoordy.T).T))
    avgData = np.mean(A_corr)
    A_corr -= avgData
    A_corr = np.reshape(A_corr, s1*s2)
    data_temp = np.zeros([s2, s1*s2])
    for i in range(s2):
        data_temp[i] = A_corr
    FT = np.matmul(data_temp * xphase, yphase.T).T
    invFT = np.fft.ifft2(np.fft.fftshift(FT)) + avgData
    return np.real(invFT)

##################################################################################
####################### Useful functions in the processing #######################
##################################################################################

#8. - sortBraggs
def sortBraggs(br, s):
    ''' Sort the Bragg peaks in the order of "lower left, lower right, upper right, and upper left" '''
    *_, s2, s1 = s
    Br_s = np.zeros_like(br)
    index_corr = [[-1,-1],[1,-1],[1,1],[-1,1]]
    center = np.array([int(s1/2),int(s2/2)])
    for i,ix in enumerate(index_corr):
        for j,jy in enumerate(np.sign(br-center)):
            if np.all(jy==ix):
                Br_s[i] = br[j]
    return Br_s

#9. - cropedge
def cropedge(A, n, obj=None, bp=None, c1=2,c2=2, a1=None, a2=None, force_commen=False, update_obj=True):
    """
    Crop out bad pixels or highly drifted regions from topo/dos map.

    Inputs:
        A           - Required : 2D or 3D array of image to be cropped.
        n           - Required : List of integers specifying how many bad pixels to crop on each side.
                                    Order: [left, right, down, up].
        obj         - Optional : Spy object of topo (2D) or map (3D).
        force_commen- Optional : Boolean determining if the atomic lattice is commensurate with
                                    the output image.

        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.

    Returns:
        A_crop  - 2D or 3D array of image after cropping.

    Usage:
        import stmpy.driftcorr as dfc
        A_crop = dfc.cropedge(A, n=5, obj=obj, corner='1')

    History:
        06/04/2019      RL : Initial commit.
        11/30/2019      RL : Add support for non-square dataset
    """
    if force_commen is not True:
        B = _rough_cut(A, n=n)
        print('Shape before crop:', end=' ')
        print(A.shape)
        print('Shape after crop:', end=' ')
        print(B.shape)
        return B
    else:
        if n != 0:
            B = _rough_cut(A, n)
        else:
            B = np.copy(A)
        *_, L2, L1 = np.shape(A)
        if bp is None:
            bp = findBraggs(A, show=False)
        bp = sortBraggs(bp, s=np.array([L2,L1]))
        bp_new = bp - np.array([int(L1/2), int(L2/2)])
        N1 = np.absolute(bp_new[0,0] - bp_new[1,0])
        N2 = np.absolute(bp_new[0,1] - bp_new[-1,1])
        offset = 0

        if a1 is None:
            a1 = c1 * L1 / N1
        if a2 is None:
            a2 = a1
            #a2 = c2 * L2 / N2
        *_, L2, L1 = np.shape(B)
        L_new1 = a1 * ((L1-offset)//(a1))
        L_new2 = a2 * ((L2-offset)//(a2))
        delta1 = (L1 - offset - L_new1) / 2
        delta2 = (L2 - offset - L_new2) / 2
        t1 = np.arange(L1)
        t2 = np.arange(L2)
        if len(np.shape(A)) == 2:
            f = interp2d(t1, t2, B, kind='cubic')
            t_new1 = np.linspace(delta1, L_new1+delta1, num=L1-offset+1)
            t_new2 = np.linspace(delta2, L_new2+delta2, num=L2-offset+1)
            z_new = f(t_new1[:-1], t_new2[:-1])
        elif len(np.shape(A)) == 3:
            z_new = np.zeros([np.shape(A)[0], L2-offset, L1-offset])
            for i in range(len(A)):
                f = interp2d(t1, t2, B[i], kind='cubic')
                t_new1 = np.linspace(0, L_new1, num=L1-offset+1)
                t_new2 = np.linspace(0, L_new2, num=L2-offset+1)
                z_new[i] = f(t_new1[:-1], t_new2[:-1])
        else:
            print('ERR: Input must be 2D or 3D numpy array!')
        if obj is not None:
            if update_obj is True:
                obj.size = obj.size * (obj.pixels - offset) / obj.pixels
                obj.pixels = obj.pixels - offset
                obj.qmag = obj.size / obj.a0
                obj.qscale = obj.pixels / (2*obj.qmag)
                obj.a1=a1
                obj.a2=a2
        return z_new

def _rough_cut(A, n):
    B = np.copy(A)
    if len(n) == 1:
        n1 = n2 = n3 = n4 = n[0]
    else:
        n1, n2, n3, n4, *_ = n
    if len(B.shape) is 2:
        if n2 == 0:
            n2 = -B.shape[1]
        if n4 == 0:
            n4 = -B.shape[0]
        return B[n3:-n4, n1:-n2]
    elif len(B.shape) is 3:
        if n2 == 0:
            n2 = -B.shape[2]
        if n4 == 0:
            n4 = -B.shape[1]
        return B[:,n3:-n4, n1:-n2]

def Gaussian2d(x, y, sigma_x, sigma_y, theta, x0, y0, Amp):
    '''
    x, y: ascending 1D array
    x0, y0: center
    '''
    a = np.cos(theta)**2/2/sigma_x**2 + np.sin(theta)**2/2/sigma_y**2
    b = -np.sin(2*theta)**2/4/sigma_x**2 + np.sin(2*theta)**2/4/sigma_y**2
    c = np.sin(theta)**2/2/sigma_x**2 + np.cos(theta)**2/2/sigma_y**2
    z = np.zeros((len(x), len(y)))
    X, Y = np.meshgrid(x, y)
    z = Amp * np.exp(-(a*(X-x0)**2 + 2*b*(X-x0)*(Y-y0) + c*(Y-y0)**2))
    return z

def FTDCfilter(A, sigma):
    '''
    Filtering DC component of Fourier transform and inverse FT, using a gaussian with one parameter sigma
    A is a 2D array, sigma is in unit of px
    '''
    *_, s2, s1 = A.shape
    m1, m2 = np.arange(s1, dtype='float'), np.arange(s2, dtype='float')
    c1, c2 = np.float((s1-1)/2), np.float((s2-1)/2)
    sigma1 = sigma
    sigma2 = sigma * s1 / s2
    g = Gaussian2d(m1, m2, sigma1, sigma2, 0, c1, c2, 1)
    ft_A = np.fft.fftshift(np.fft.fft2(A))
    ft_Af = ft_A * g
    Af = np.fft.ifft2(np.fft.ifftshift(ft_Af))
    return np.real(Af)

def unwrap_phase_2d(A, thres=None):
    output = np.copy(A[::-1,::-1])
    if len(np.shape(A)) == 2:
        n = np.shape(A)[-1]
        for i in range(n):
            output[i,:] = unwrap_phase(output[i,:], tolerance=thres)
        for i in range(n):
            output[:,i] = unwrap_phase(output[:,i], tolerance=thres)
        return output[::-1,::-1]

#10. - compute_dist
def compute_dist(x1, x2, p=None):

    if p is None:
        p1, p2 = 1, 1
    else:
        p1, p2 = p
    return np.sqrt(((x1[0]-x2[0])*p1)**2+((x1[1]-x2[1])*p2)**2)

#11. - global_corr
def global_corr(A, obj=None, bp=None, show=False, angle=np.pi/4, update_obj=True, **kwargs):
    """
    Global shear correct the 2D topo automatically.

    Inputs:
        A           - Required : 2D array of topo to be shear corrected.
        obj         - Optional : Spy object of topo (2D) or map (3D).
        bp          - Optional : Bragg points. If not offered, it will calculated from findBraggs(A)
        show        - Optional : Boolean specifying if the results are plotted or not
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.
        angle       - Optional : orientation of the Bragg peaks, default as pi/4. Will be passed to gshearcorr
        **kwargs    - Optional : key word arguments for findBraggs function
    Returns:
        bp_1    - Bragg peaks returned by gshearcorr. To be used in local_corr()
        data_1  - 2D array of topo after global shear correction

    Usage:
        import stmpy.driftcorr as dfc
        bp1, data1 = dfc.global_corr(A, show=True)

    History:
        04/29/2019      RL : Initial commit.
    """
    *_, s2, s1 = np.shape(A)
    if bp is None:
        bp_1 = findBraggs(A, obj=obj, thres=0.2, show=show, update_obj=update_obj, **kwargs)
    else:
        bp_1 = bp
    m, data_1 = gshearcorr(A, bp_1, obj=obj, rspace=True, angle=angle, update_obj=update_obj)
    if show is True:
        fig,ax=plt.subplots(1,2,figsize=[8,4])
        ax[0].imshow(data_1, cmap=stmpy.cm.blue2,origin='lower')
        ax[0].set_xlim(0,s1)
        ax[0].set_ylim(0,s2)
        ax[1].imshow(stmpy.tools.fft(data_1, zeroDC=True), cmap=stmpy.cm.gray_r,origin='lower')
        fig.suptitle('After global shear correction', fontsize=14)
        fig,ax=plt.subplots(2,2,figsize=[8,8])
        ax[0,0].imshow(data_1, cmap=stmpy.cm.blue2,origin='lower')
        ax[0,1].imshow(data_1, cmap=stmpy.cm.blue2,origin='lower')
        ax[1,0].imshow(data_1, cmap=stmpy.cm.blue2,origin='lower')
        ax[1,1].imshow(data_1, cmap=stmpy.cm.blue2,origin='lower')
        ax[0,0].set_xlim(0, s1/10)
        ax[0,0].set_ylim(s2-s2/10, s2)
        ax[0,1].set_xlim(s1-s1/10, s1)
        ax[0,1].set_ylim(s2-s2/10, s2)
        ax[1,0].set_xlim(0, s1/10)
        ax[1,0].set_ylim(0, s2/10)
        ax[1,1].set_xlim(s1-s1/10, s1)
        ax[1,1].set_ylim(0, s2/10)
        fig.suptitle('Bad pixels in 4 corners', fontsize=14)
    return m, data_1

#12. - local_corr
def local_corr(data, obj=None, bp=None, sigma=10, method="lockin", fixMethod='unwrap', show=False, update_obj=True, **kwargs):
    """
    Locally drift correct 2D topo automatically.

    Inputs:
        data        - Required : 2D array of topo after global shear correction, with bad pixels removed on the edge
        sigma       - Optional : Floating number specifying the size of mask to be used in phasemap()
        obj         - Optional : Spy object of topo (2D) or map (3D).
        bp          - Optional : Bragg points. If not offered, it will calculated from findBraggs(A)
        method      - Optional : Specifying which method to in phasemap()
                                "lockin": Spatial lock-in method to find phase map
                                "convolution": Gaussian mask convolution method to find phase map
        fixMethod   - Optional : Specifying which method to use in fixphaseslip()
                                "unwrap": fix phase jumps line by line in x direction and y direction, respectively
                                "spiral": fix phase slip in phase shift maps by flattening A into a 1D array in a spiral way
        show        - Optional : Boolean specifying if the results are plotted or not
        update_obj  - Optional : Boolean, if True then all the attributes of the object will be updated.
        **kwargs    - Optional : key word arguments for findBraggs function

    Returns:
        ux          - 2D array of drift field in x direction
        uy          - 2D array of drift field in y direction
        data_corr   - 2D array of topo after local drift corrected

    Usage:
        import stmpy.driftcorr as dfc
        ux, uy, data_corr = dfc.local_corr(data, sigma=5, method='lockin', fixMethod='unwrap', show=True)

    History:
        04/29/2019      RL : Initial commit.
    """
    *_, s2, s1 = np.shape(data)
    if bp is None:
        bp_2 = findBraggs(data, obj=obj, thres=0.2, show=show, update_obj=update_obj, **kwargs)
    else:
        bp_2 = bp
    thetax, thetay, Q1, Q2= phasemap(data, bp=bp_2, obj=obj, method=method, sigma=sigma, update_obj=update_obj)
    if show is True:
        fig,ax=plt.subplots(1,2,figsize=[8,4])
        ax[0].imshow(thetax, origin='lower')
        ax[1].imshow(thetay, origin='lower')
        fig.suptitle('Raw phase maps')
    thetaxf = fixphaseslip(thetax, method=fixMethod)
    thetayf = fixphaseslip(thetay, method=fixMethod)
    if show is True:
        fig,ax=plt.subplots(1,2,figsize=[8,4])
        ax[0].imshow(thetaxf, origin='lower')
        ax[1].imshow(thetayf, origin='lower')
        fig.suptitle('After fixing phase slips')
    ux, uy = driftmap(thetaxf, thetayf, Q1, Q2, obj=obj, method=method, update_obj=update_obj)
    if method=='lockin':
        data_corr = driftcorr(data, ux, uy, method='lockin', interpolation='cubic')
    elif method=='convolution':
        data_corr = driftcorr(data, ux, uy, method='convolution',)
    else:
        print("Error: Only two methods are available, lockin or convolution.")
    if show is True:
        fig,ax=plt.subplots(2,2,figsize=[8, 8])
        ax[1, 0].imshow(data_corr, cmap=stmpy.cm.blue1,origin='lower')
        ax[1, 1].imshow(stmpy.tools.fft(data_corr, zeroDC=True), cmap=stmpy.cm.gray_r,origin='lower')
        ax[0, 0].imshow(data, cmap=stmpy.cm.blue1,origin='lower')
        ax[0, 1].imshow(stmpy.tools.fft(data, zeroDC=True), cmap=stmpy.cm.gray_r,origin='lower')
        fig.suptitle('Before and after local drift correction')
    return ux, uy, data_corr

#14. - apply_dfc_3d
def apply_dfc_3d(data, ux, uy, matrix, bp=None, obj=None, n1=None, n2=None, method='convolution',update_obj=False):
    """
    Apply drift field (both global and local) found in 2D to corresponding 3D map.

    Inputs:
        data        - Required : 3D array of map to be drift corrected
        bp         - Required : Coordinates of Bragg peaks returned by local_corr()
        ux          - Required : 2D array of drift field in x direction. Usually generated by local_corr()
        uy          - Required : 2D array of drift field in y direction. Usually generated by local_corr()
        crop1       - Optional : List of length 1 or length 4, specifying after global shear correction how much to crop on the edge
        crop2       - Optional : List of length 1 or length 4, specifying after local drift correction how much to crop on the edge
        method      - Optional : Specifying which method to apply the drift correction
                                    "interpolate": Interpolate A and then apply it to a new set of coordinates,
                                                    (x-ux, y-uy)
                                    "invfft": Used inversion fft to apply the drift fields
    Returns:
        data_corr   - 2D array of topo after local drift corrected

    Usage:
        import stmpy.driftcorr as dfc
        data_corr = dfc.apply_dfc_3d(data, bp=bp, ux=ux, uy=uy, crop1=[5], crop2=[5], method='interpolate')

    History:
        04/29/2019      RL : Initial commit.
    """
    data_c = np.zeros_like(data)
    for i in range(len(data)):
        _, data_c[i]  = gshearcorr(data[i], matrix=matrix, obj=obj, rspace=True, update_obj=update_obj)
    if n1 is None:
        data_c = data_c
    else:
        data_c = cropedge(data_c, n=n1)
    data_corr = driftcorr(data_c, ux, uy, method=method, interpolation='cubic')
    if n2 is None:
        data_out = data_corr
    else:
        data_out = cropedge(data_corr, obj=obj, bp=bp, n=n2, force_commen=True, update_obj=update_obj)
    return data_out

#15. - display
def display(A, B=None, sigma=3, clim_same=True):
    '''
    Display or compare images in both real space and q-space.

    Inputs:
        A           - Required : Real space image to display.
        B           - Optional : Another real space image to be compared with A.
        sigma       - Optional : sigma for the color limit.
        clim_same   - Optional : If True, then both FT of A and B will be displayed under the
                                    same color limit (determined by A).

    Returns:
        N/A

    Usage:
        import stmpy.driftcorr as dfc
        dfc.display(topo.z)

    '''
    if B is None:
        A_fft = stmpy.tools.fft(A, zeroDC=True)
        c = np.mean(A_fft)
        s = np.std(A_fft)
        fig,ax=plt.subplots(1,2,figsize=[8, 4])
        ax[0].imshow(A, cmap=stmpy.cm.blue2, origin='lower')
        ax[1].imshow(A_fft, cmap=stmpy.cm.gray_r, origin='lower', clim=[0,c+sigma*s])
    else:
        A_fft = stmpy.tools.fft(A, zeroDC=True)
        B_fft = stmpy.tools.fft(B, zeroDC=True)
        c1 = np.mean(A_fft)
        s1 = np.std(A_fft)
        if clim_same is True:
            c2 = c1
            s2 = s1
        else:
            c2 = np.mean(B_fft)
            s2 = np.std(B_fft)
        fig,ax=plt.subplots(2,2,figsize=[8, 8])
        ax[0,0].imshow(A, cmap=stmpy.cm.blue2, origin='lower')
        ax[0,1].imshow(A_fft, cmap=stmpy.cm.gray_r, origin='lower', clim=[0,c1+sigma*s1])
        ax[1,0].imshow(B, cmap=stmpy.cm.blue2, origin='lower')
        ax[1,1].imshow(B_fft, cmap=stmpy.cm.gray_r, origin='lower', clim=[0,c2+sigma*s2])
        for axis in ax.flatten():
            axis.set_aspect(1)
        plt.tight_layout()


def quick_linecut(A, width=2, n=4, bp=None, ax=None, thres=3):
    """
    Take four linecuts automatically, horizontal, vertical, and two diagonal.
    Inputs:
        A           - Required : FT space image to take linecuts.
        width       - Optional : Number of pixels for averaging.
        bp          - Optional : Bragg peaks
        thres       - Optional : threshold for displaying FT

    Returns:
        N/A

    Usage:
        import stmpy.driftcorr as dfc
        r, cut = dfc.quick_linecut(A)

    """
    Y = np.shape(A)[-2] / 2
    X = np.shape(A)[-1] / 2
    r = []
    cut = []
    start = [[0,Y],[X,0],[0,0],[0,Y*2]]
    end = [[X*2, Y],[X, Y*2],[X*2, Y*2], [X*2, 0]]
    color = ['r','g','b','k']

    plt.figure(figsize=[4,4])
    if len(np.shape(A)) == 3:
        if bp is None:
            bp_x = np.min(findBraggs(np.mean(A, axis=0), rspace=False))
        else:
            bp_x = bp
        cm = np.mean(np.mean(A, axis=0))
        cs = np.std(np.mean(A, axis=0))
        plt.imshow(np.mean(A, axis=0), clim=[0,cm+thres*cs])
    elif len(np.shape(A)) == 2:
        if bp is None:
            bp_x = np.min(findBraggs(A, rspace=False))
        else:
            bp_x = bp
        cm = np.mean(A)
        cs = np.std(A)
        plt.imshow(A, clim=[0, cm+thres*cs])

    qscale = X*2 / (X*2 - bp_x * 2)

    for i in range(n):
        r1, cut1 = stmpy.tools.linecut(A, start[i], end[i],
            width = width, show=True, ax=plt.gca(), color=color[i])
        r.append(r1)
        cut.append(cut1)
    plt.gca().set_xlim(-1, X*2+1)
    plt.gca().set_ylim(-1, Y*2+1)
    return qscale, cut

def quick_show(A, en, thres=5, rspace=True, saveon=False, qlimit=1.2, imgName='', extension='png'):
    layers = len(A)
    if rspace is False:
        imgsize = np.shape(A)[-1]
        bp_x = np.min(findBraggs(np.mean(A, axis=0), min_dist=int(imgsize/10), rspace=rspace))
        ext = imgsize / (imgsize - 2*bp_x)
    if layers > 12:
        skip = layers // 12
    else:
        skip = 1
    fig,ax=plt.subplots(3,4,figsize=[16,12])
    try:
        for i in range(12):
            c = np.mean(A[i*skip])
            s = np.std(A[i*skip])
            if rspace is True:
                ax[i//4,i%4].imshow(A[i*skip], clim=[c-thres*s,c+thres*s],cmap=stmpy.cm.jackyPSD)
            else:
                ax[i//4,i%4].imshow(A[i*skip],extent=[-ext,ext,-ext,ext,],clim=[0,c+thres*s],cmap=stmpy.cm.gray_r)
                ax[i//4,i%4].set_xlim(-qlimit,qlimit)
                ax[i//4,i%4].set_ylim(-qlimit,qlimit)
            ax[i//4,i%4].set_aspect(1)
            stmpy.image.add_label("${}$ mV".format(int(en[i*skip])), ax=ax[i//4,i%4])
        plt.tight_layout()
    except IndexError:
        pass
    if saveon is True:
        plt.savefig("{}.{}".format(imgName, extension), bbox_inches='tight')

def quick_show_cut(A, en, qscale, thres=5, thres2=None, saveon=False, qlimit=1.2, imgName='', extension="png"):
    fname = ["M-0", "M-90", "X-45","X-135"]
    X1, Y1 = np.shape(A[0])
    X2, Y2 = np.shape(A[-1])
    q1 = np.linspace(-qscale, qscale, num=Y1)
    q2 = np.linspace(-qscale*np.sqrt(2), qscale*np.sqrt(2), num=Y2)
    if thres2 is None:
        thres2 = thres
    for i,ix in enumerate(A):
        plt.figure(figsize=[6,3])
        c = np.mean(ix)
        s = np.std(ix)
        if i in [0,1]:
            plt.pcolormesh(q1, en, ix,cmap=stmpy.cm.gray_r, vmin=0, vmax=c+thres*s)
        else:
            plt.pcolormesh(q2, en, ix,cmap=stmpy.cm.gray_r, vmin=0, vmax=c+thres2*s)
        plt.gca().set_xlim(-qlimit,qlimit)
        plt.axvline(-1, linestyle='--')
        plt.axvline(1, linestyle='--')
        if saveon is True:
            plt.savefig(imgName + " along {}.{}".format(fname[i], extension), facecolor='w')

# Quick show single images
def quick_show_single(A, en, thres=5, qscale=None, rspace=False, saveon=False, qlimit=1.2, imgName='', extension='png'):
    layers = len(A)
    if rspace is False:
        if qscale is None:
            imgsize = np.shape(A)[-1]
            if len(np.shape(A)) == 3:
                A_topo = np.mean(A, axis=0)
            else:
                A_topo = A
            bp_x = np.min(findBraggs(A_topo, min_dist=int(imgsize/10), rspace=rspace))
            ext = imgsize / (imgsize - 2*bp_x)
        else:
            ext = qscale
    if len(np.shape(A)) == 3:
        for i in range(layers):
            plt.figure(figsize=[5,5])
            c = np.mean(A[i])
            s = np.std(A[i])
            if rspace is True:
                plt.imshow(A[i], clim=[c-thres*s,c+thres*s],cmap=stmpy.cm.jackyPSD)
            else:
                plt.imshow(A[i],extent=[-ext,ext,-ext,ext,],clim=[0,c+thres*s],cmap=stmpy.cm.gray_r)
                plt.xlim(-qlimit,qlimit)
                plt.ylim(-qlimit,qlimit)
            stmpy.image.add_label("${}$ mV".format(int(en[i])), ax=plt.gca())
            plt.gca().axes.get_xaxis().set_visible(False)
            plt.gca().axes.get_yaxis().set_visible(False)
            plt.gca().set_frame_on(False)
            plt.gca().set_aspect(1)
            if saveon is True:
                plt.savefig("{} at {}mV.{}".format(imgName, int(en[i]), extension), dpi=400, bbox_inches='tight',pad_inches=0)
    elif len(np.shape(A)) == 2:
        plt.figure(figsize=[5,5])
        c = np.mean(A)
        s = np.std(A)
        if rspace is True:
            plt.imshow(A, clim=[c-thres*s,c+thres*s],cmap=stmpy.cm.jackyPSD)
        else:
            plt.imshow(A,extent=[-ext,ext,-ext,ext,],clim=[0,c+thres*s],cmap=stmpy.cm.gray_r)
            plt.xlim(-qlimit,qlimit)
            plt.ylim(-qlimit,qlimit)
        stmpy.image.add_label("${}$ mV".format(int(en)), ax=plt.gca())
        plt.gca().axes.get_xaxis().set_visible(False)
        plt.gca().axes.get_yaxis().set_visible(False)
        plt.gca().set_frame_on(False)
        plt.gca().set_aspect(1)
        if saveon is True:
            plt.savefig("{} at {}mV.{}".format(imgName, int(en), extension), dpi=400, bbox_inches='tight',pad_inches=0)



import scipy.io as sio

def load_STMView(filePath):
    '''Load an STMView file into python as an object.
        Usage: data = load(filePath)
        '''
    fileObject = sio.loadmat(filePath)
    typeID = []
    dataObjects = {}
    for dataKey in fileObject.keys():
        if dataKey not in ['__header__','__version__','__globals__']:
            typeID.append(dataKey)
    for ID in typeID:
        dataObjects[ID] = pyData(fileObject,ID)
    return dataObjects

def save(filePath,pyObject):
    '''Save a python object into STMView format. 
        Under deveopment - anticipate some bugs.
        Usage: save(filePath, dataObject)
        '''
    pySave = {}
    for key in pyObject.keys():
        pyDict = vars(pyObject[key])
        pyDict['info'] = vars(pyObject[key].info)
        pySave[key] = pyDict
    sio.savemat(filePath, pySave)

class pyData(object):
    def __init__(self,fileObject,typeID):
        contents = fileObject[typeID]
        for name in contents.dtype.names:
            setattr(self, name, contents[name][0,0])
        self.info = pyInfo(self.info)

class pyInfo(object):
    def __init__(self,info) :
        for name in info.dtype.names:
             setattr(self, name, info[name][0,0])




from setuptools import setup

setup(
  name = 'getstmpy',
  packages = ['stmpy'],
  version = '0.1',
  license='MIT',
  description = 'Scanning tunneling microscopy data analysis suite',
  author = 'Harris Pirie',
  author_email = 'hpirie@live.com',
  url = 'https://github.com/harrispirie/stmpy',
  download_url = 'https://github.com/harrispirie/stmpy/archive/v0.1.tar.gz',
  keywords = ['STM', 'Python', 'Data Analysis'],
  install_requires=[
          'numpy',
          'scipy',
          'matplotlib',
          'opencv-python',
          'scikit-image'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
  ],
)



import numpy as np
import pylab as plt
import matplotlib as mpl
from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable, inset_locator
from matplotlib import cm
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import matplotlib.font_manager as fm

boxProperties = dict(boxstyle='square', facecolor='w', alpha=0.7, linewidth=0.0)
textOptions = dict(fontsize=12, color = 'k', bbox=boxProperties, ha='right', va='top')

def saturate(level_low=0, level_high=None, im=None):
    '''
    Adjusts color axis of in current handle.  Calculates a probablility density
    function for the data in current axes handle.  Uses upper and lower
    thresholds to find sensible c-axis limits.  Thresholds are between 0 and
    100.  If unspecified the upper threshold is assumed to be 100 - lower
    threshold.

    Usage:  pcolormesh(image)
            saturate(10)
    '''
    level_low = float(level_low) / 200.0
    if level_high is None:
        level_high = 1-level_low
    else:
        level_high = (float(level_high)+100) / 200.0
    if im is not None:
        images = [im]
        data = im.get_array().ravel()
    else:
        imageObjects = mpl.pyplot.gca().get_children()
        data = []
        images = []
        for item in imageObjects:
            if isinstance(item, (mpl.image.AxesImage, mpl.collections.QuadMesh)):
                images.append(item)
                data.append(item.get_array().ravel())
    y = sorted(np.array(data).ravel())
    y_density = np.absolute(y) / sum(np.absolute(y))
    pdf = np.cumsum(y_density)
    y_low = np.absolute(level_low - pdf)
    y_high = np.absolute(level_high - pdf)
    c_low = y[np.argmin(y_low)]
    c_high = y[np.argmin(y_high)]
    for image in images:
        image.set_clim(c_low, c_high)


def write_animation(data, fileName, saturation=2, clims=(0,1), cmap=None,
                    label=None, label_caption='meV', speed=8, zoom=1, **kwargs):
    ''' Create a movie from a 3D data set and save it to the path specified. Intended
    for visualising DOS maps and QPI data. Iterates through the first index in
    the data set to create an animation.

    Notes:
        Make sure you include file extension (eg '.mov') in the file path.
        Currently supported file types are mov and mp4.
        Make sure you have ffmpeg installed (e.g. through Homebrew) before running.
        The MP4 writer is really difficult to get high quality.  I learned that
        it depends on figsize, bitrate and dpi.  Basically, if the combination
        of these parameters falls outside of a narrow window the compression
        session will fail to start.  The defaults used in this script seem to
        work.


    Inputs:
        data    - Required : A 3D numpy array containing each frame in the
                             movie.
        fileName - Required : Full path for output file including extension.
        saturation - Optional : Saturation applied to each frame separately.
        clims   - Optional : If saturation=None, user defined clims are used.
                             Specify as a tuple (cmin, cmax).
        cmap    - Optional : Colormap instance will default to cm.bone_r if not
                             specified.
        label   - Optional : Data to use for labeling frames.  Must have the
                             same length as data.shape[0].
        label_caption - Optional : String to be displayed after label.
        speed   - Optional : Frames per second for output.
        zoom    - Optional : Float for zoom factor (eg. 2 times zoom).
        **kwarg - Optional : Additional keyword arguments are sent to imshow().

    Returns:
        Has no returns.

    Usage:
        write_animation(data, fileName, saturation=2, clims=(0,1), cmap=None,
                    label=None, label_caption='meV', speed=8, zoom=1, **kwargs)

    History:
        2017-06-08  -   HP  : Added **kwargs sent to imshow
        2017-06-23  -   HP  : Added support for MP4 export and made codec
                              choice automatic.
    '''
    boxProperties = dict(boxstyle='square', facecolor='w', alpha=0.8, linewidth=0.0)
    textOptions = dict(fontsize=14, color='k', bbox=boxProperties, ha='right', va='center')

    if cmap is None:
        cmap = cm.bone_r
    fig = plt.figure(figsize=[4,4])
    ax = plt.subplot(111)
    ax.set_xticks([]), ax.set_yticks([])
    im = plt.imshow(data[0], cmap=cmap, extent=[-1,1,-1,1], origin='lower',
                    **kwargs)
    ax.set_xlim(-1.0/zoom, 1.0/zoom)
    ax.set_ylim(-1.0/zoom, 1.0/zoom)
    if saturation is not None:
        saturate(saturation, im=im)
    else:
        plt.clim(clims)

    if label is not None:
        tx = ax.text(0.95,0.95,'{:2.1f} {:}'.format(label[0], label_caption),
                  transform=ax.transAxes, **textOptions)
    def init():
        im.set_array(data[0])
        ax.text(20,200,'')
        return [im]

    def animate(i):
        im.set_array(data[i])
        if saturation is not None:
            saturate(saturation, im=im)
        else:
            plt.clim(clims)
        if label is not None:
            tx.set_text('{:2.1f} {:}'.format(label[i], label_caption))
        return [im]
    fig.tight_layout()
    ani = FuncAnimation(fig, animate, init_func=init, frames=data.shape[0])
    if fileName.endswith('.mov'):
        ani.save(fileName, codec='prores', dpi=200, fps=speed)
    elif fileName.endswith('.mp4'):
        ani.save(fileName, dpi=200, bitrate=1e5, fps=speed)
    else:
        print('ERR: fileName must end with .mov or .mp4')


def add_colorbar(loc=0, label='', fs=12, size='5%', pad=0.05, ax=None, im=None,
        ticks=True):
    '''Add a colorbar to the current axis. Alternatively, use add_cbar().

    Inputs:
        loc     - Optional : Specify the location of the colorbar: 0 (bottom),
                             1 (right), 2 (top) or 3 (left).
        label   - Optional : String containing colorbar label
        fs      - Optional : Float for colorbar label fontsize
        size    - Optional : String describing the width of the colorbar as a
                             percentage fo the image.
        pad     - Optional : Float for colorbar pad from image axes.
        ax      - Optional : Axes to attach the colorbar to.  Uses gca() as
                             default.
        im      - Optional : Image used to get colormap and color limits.

    Returns:
        cbar    - matplotlib.colorbar.Colorbar instance.

    History:
        2017-07-20  - HP : Initial commit.
        2017-08-06  - HP : Added flag for removing ticks.
    '''
    if ax is None:
        ax = mpl.pyplot.gca()
    if im is None:
        elements = ax.get_children()
        for element in elements:
            if isinstance(element, (mpl.image.AxesImage, mpl.collections.QuadMesh)):
                im = element
    divider = make_axes_locatable(ax)
    fig = ax.get_figure()
    if loc == 0:
        cax = divider.new_vertical(size=size, pad=pad, pack_start=True)
        fig.add_axes(cax)
        cbar = fig.colorbar(im, cax=cax, orientation='horizontal')
        cbar.set_label(label, fontsize=fs)
    elif loc == 1:
        cax = divider.new_horizontal(size=size, pad=pad, pack_start=False)
        fig.add_axes(cax)
        cbar = fig.colorbar(im, cax=cax, orientation='vertical')
        cbar.set_label(label, fontsize=fs)
    elif loc == 2:
        cax = divider.new_vertical(size=size, pad=pad, pack_start=False)
        fig.add_axes(cax)
        cbar = fig.colorbar(im, cax=cax, orientation='horizontal')
        cbar.set_label(label, fontsize=fs)
        cbar.ax.xaxis.set_ticks_position('top')
        cbar.ax.xaxis.set_label_position('top')
    elif loc == 3:
        cax = divider.new_horizontal(size=size, pad=pad, pack_start=True)
        fig.add_axes(cax)
        cbar = fig.colorbar(im, cax=cax, orientation='vertical')
        cbar.set_label(label, fontsize=fs)
        cbar.ax.yaxis.set_ticks_position('left')
        cbar.ax.yaxis.set_label_position('left')
    else:
        raise ValueError('loc must be 0 (bottom), 1 (right), 2 (top) or 3 (left).')
    if ticks is False:
        cbar.set_ticks([])
    return cbar

def add_cbar(ax=None, im=None, orient='v', length='45%', thickness='7%',
        hPos=None, vPos=None, sf=2, units='', fs=17, pad=0.12, labelpad=0.05):
    '''Adds colorbar to current axis.

    Inputs:
        ax      - Optional : Axes to attach the colorbar to.  Uses gca() as
                             default.
        im      - Optional : Image used to get colormap and color limits.
        orient  - Optional : 'v' for vertical or 'h' for horizontal.
                             Default: 'v'
        length  - Optional : String containing colorbar length as a percent.
                             Default: '45%' (of ax)
        thickness - Optional : String containing colorbar thickness as a percent.
                             Default: '7%' (of ax)
        hPos    - Optional : Float. Horizontal position from lower left of ax.
                             Default: 1.1 (for vertical) (ie 10% away from RHS)
        vPos    - Optional : Float. Vertical position from lower left of ax.
                             Default: 0.1 (for vertical) (ie 10% away from BOTTOM)
        sf      - Optional : Int. Number of significant figures for tick labels.
                             Default: 2
        units   - Optional : String to put after tick label.
                             Default: ''
        fs      - Optional : Float for label fontsize.
                             Default: 17
        pad     - Optional : Float that pushes the colorbar away from ax.
                             Default: 0.12 (equivalent to 12% of ax)
        labelpad - Optional : Float. Distance from edge of colorbar to start of label.
                              Default: 0.05 (equivalent to 5% of ax)

    Returns:
        cbar    - matplotlib.colorbar.Colorbar instance.

    History:
        2019-11-02  - HP : Initial commit.
        2019-11-24  - HP : Added horizontal compatibility.
    '''
    fm ='%.' + str(sf) + 'f '
    ln = float(length.split('%')[0])/100
    th = float(thickness.split('%')[0])/100
    if ax is None:
        ax = mpl.pyplot.gca()
    if im is None:
        elements = ax.get_children()
        for element in elements:
            if isinstance(element, (mpl.image.AxesImage, mpl.collections.QuadMesh)):
                im = element
    if orient == 'v':
        if hPos is None:
            hPos = 1
        if vPos is None:
            vPos = 0.12
        hPos += pad
        axins = inset_locator.inset_axes(ax, width=thickness, height=length,
                        loc='lower left', bbox_to_anchor=(hPos, vPos, 1, 1),
                        bbox_transform=ax.transAxes, borderpad=0)
        cb = mpl.pyplot.colorbar(im, cax=axins, orientation='vertical')
        low, high = cb.get_clim()
        axins.text(hPos+th/2, vPos+ln+labelpad, fm % high + units,
                transform=ax.transAxes, fontsize=fs, ha='center')
        axins.text(hPos+th/2, vPos-labelpad, fm % low + units,
                transform=ax.transAxes, fontsize=fs, ha='center', va='top')
    elif orient == 'h':
        if hPos is None:
            hPos = 0.5 - ln/2
        if vPos is None:
            vPos = 0
        vPos -= pad
        axins = inset_locator.inset_axes(ax, width=length, height=thickness,
                        loc='lower left', bbox_to_anchor=(hPos, vPos, 1, 1),
                        bbox_transform=ax.transAxes, borderpad=0)
        cb = mpl.pyplot.colorbar(im, cax=axins, orientation='horizontal')
        low, high = cb.get_clim()
        axins.text(hPos-labelpad, vPos+th/2, fm % low + units,
                transform=ax.transAxes, fontsize=fs, ha='right', va='center')
        axins.text(hPos+ln+labelpad, vPos+th/2, fm % high + units,
                transform=ax.transAxes, fontsize=fs, ha='left', va='center')
    cb.set_ticks([])
    return cb


def add_label(label, loc=0, ax=None, fs=20, txOptions=None, bbox=None):
    '''Add text labels to images.

    Inputs:
        label   - Required : String (or formatted string) with contents of
                             label.
        loc     - Optional : Integer to describe location of label:
                                 0: top-right
                                 1: top-left
                                 2: bottom-left
                                 3: bottom-right
        ax      - Optional : Axes to place the label.  Uses mpl.pyplot.gca() if
                            not provided.
        txOptions - Optional : Dictionary containing keyword arguments for text
                               formatting, such as: 'fontsize' and 'color'.
        bbox    - Optional : Dictionary containing box properties, such as:
                             'boxstyle', 'facecolor', 'alpha', 'linewidth'.  If
                             no box is desired set to False.

    Returns:
        tx  -   mpl.pyplot.text opbject.

    History:
        2017-08-15  - HP : Initial commit.
        2017-08-31  - HP : Added new locations for text.
    '''
    if ax is None:
        ax = mpl.pyplot.gca()
    if bbox is None:
        bbox = dict(boxstyle='square', facecolor='w', alpha=0.8, linewidth=0.0)
        if txOptions is None:
            txOptions = dict(fontsize=fs, color='k', bbox=bbox)
    elif bbox is False:
        if txOptions is None:
            txOptions = dict(fontsize=fs, color='k')
    else:
        if txOptions is None:
            txOptions = dict(fontsize=fs, color='k', bbox=bbox)

    if loc == 0:
        tx = ax.text(0.95,0.95, label, va='top', ha='right',
                    transform=ax.transAxes, **txOptions)
    elif loc == 1:
        tx = ax.text(0.05,0.95, label, va='top', ha='left',
                    transform=ax.transAxes, **txOptions)
    elif loc == 2:
        tx = ax.text(0.05,0.05, label, va='bottom', ha='left',
                    transform=ax.transAxes, **txOptions)
    elif loc == 3:
        tx = ax.text(0.95,0.05, label, va='bottom', ha='right',
                    transform=ax.transAxes, **txOptions)
    else:
        raise ValueError('loc must be one of the following options:\n' +
                    '\t\t0: top-right\n\t\t1: top-left\n\t\t2: bottom-left' +
                    '\n\t\t3:bottom-right')

    return tx

def add_scale_bar(length, imgsize, imgpixels, barheight=1.5e-2, axes=None, unit='nm',
        loc='lower right', color='w', fs=20, pad=0.5, tex=True, **kwargs):
    """
    Add scale bar to images.

    Inputs:
        length      - Required : Float. The length of scale bar in the given
                                unit.
        imgsize     - Required : Float. The size of image in the given unit.
        imgpixels   - Required : Float. The size of image in unit of pixels.
        barheight   - Optional : Float. The ratio between height of scale bar
                                and the height of the whole image. Default: 1.5 %.
        axes        - Optional : Axes object. It controls which axis the scale
                                bar is added to. If not given, gca() will be the
                                default value.
        unit        - Optional : String. Unit that will be shown in the text of
                                scale bar. Default: nm
        loc         - Optional : Int or string. Default: 'lower right'.
                                'best'          0
                                'upper right'   1
                                'upper left'    2
                                'lower left'    3
                                'lower right'   4
                                'right'         5
                                'center left'   6
                                'center right'  7
                                'lower center'  8
                                'upper center'  9
                                'center'        10
        color       - Optional : String. It specifies the color of the scale bar
                                and the text on it.
        fs          - Optional : Float. Fontsize for the label.
        pad         - Optional : Float. Distance from edge of image.
        tex         - Optional : Boolean. If false, the font is taken from
                                 rcparams. (If true it uses latex fonts).
        **kwargs    - Optional : Passed to AnchorSizeBar(). Including:
                                 'sep' : separation between bar and label
                                 'label_top' : top put label above bar

    Returns:
        scalebar    - scalebar object, which can be added to axis object Ax with
                                "Ax.add_artist(scalebar)"

    History:
        2018-11-02  - RL : Initial commit.
        2019-10-22  - HP : Add option to *not* use tex formatting!
    """

    if axes is None:
        axes = gca()
    to_distance = imgsize/imgpixels
    fontprops = fm.FontProperties(size=fs)
    if tex:
        scalebar = AnchoredSizeBar(axes.transData,
            length/to_distance, r'$\mathbf{{{}''\ {}}}$'.format(length, unit), loc,
            pad=pad,
            color=color,
            frameon=False,
            size_vertical=int(imgpixels*barheight),
            fontproperties=fontprops,
            **kwargs
            )
    else:
        scalebar = AnchoredSizeBar(axes.transData,
            length/to_distance, '{} {}'.format(length, unit), loc,
            pad=pad,
            color=color,
            frameon=False,
            size_vertical=int(imgpixels*barheight),
            fontproperties=fontprops,
            **kwargs
            )

    axes.add_artist(scalebar)
    return scalebar

def get_cbar(cmap, figsize=[0.36,0.07], color='k', orientation='horizontal',
        nseg=100):
    """
    Plot a basic colorbar for a given colormap.

    Inputs:
        cmap    - Required : Instance of stmpy.cm.
        figsize - Optional : Desired size for the colorbar
        color   - Optional : Border color
        orientation - Optional : 'horizontal' or 'vertical'
        nseg    - Optional : Number of discrete colors to use in plotting.

    Returns:
        cb - matplotlib colorbar instance.

    History:
    2020-04-25  - HP : Initial commit.
    """
    a = np.array([[0,1]])
    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111)
    im = ax.imshow(a, cmap=cmap)
    ax.set_visible(False)
    cax = fig.add_axes([0.05, 0.05, 0.9, 0.9])
    bounds = np.linspace(0, 1, nseg+1)
    cb = plt.colorbar(im, cax=cax, orientation=orientation, boundaries=bounds)
    cb.set_ticks([])
    cb.outline.set_edgecolor(color)
    return cb
